/*
 *  dns.c (DNS̃LbV֘ÃW[)
 *
 *  Writen by Masaru Tsuchiyama <m.tsutsu@gmail.com>
 */
#include <windows.h>
#include <stdio.h>
#include <stdarg.h>

void *Malloc(int size);
char * DupStr(char *buf);

struct DNS_CACHE{
	char *szHost;
	unsigned long addr;
};

#define DNS_CACHE_SIZE	100 // LbVp̃f[^̊mےP

static CRITICAL_SECTION g_csDNS_Cache;	// LbVւ̔rANZXp
static struct DNS_CACHE *g_dns_cache;	// LbVf[^{
static int g_nCacheCount;	// ۂɊi[ĂTCYTCY
static int g_nCacheSize;	// LbVpɊmۂĂTCY

/*
 *	LbV֐B
 */
void InitializeDNSCache(void)
{
	if(g_dns_cache == NULL){
		InitializeCriticalSection(&g_csDNS_Cache);
		g_dns_cache = Malloc(DNS_CACHE_SIZE * sizeof(struct DNS_CACHE));
		EnterCriticalSection(&g_csDNS_Cache);
		g_nCacheCount = 0;
		g_nCacheSize  = DNS_CACHE_SIZE;
		LeaveCriticalSection(&g_csDNS_Cache);
	}
}

/*
 *	LbVNA֐B
 */
void ClearDNSCache(void)
{
	int i;
	InitializeDNSCache();
	EnterCriticalSection(&g_csDNS_Cache);
	if(g_dns_cache){
		for(i=0;i<g_nCacheCount;i++){
			GlobalFree(g_dns_cache[i].szHost);
			g_dns_cache[i].szHost = NULL;
		}
		g_nCacheCount = 0;
		GlobalFree(g_dns_cache);
		g_dns_cache = NULL;
	}
	LeaveCriticalSection(&g_csDNS_Cache);
}

/*
 *	LbVj֐B
 */
void DestroyDNSCache(void)
{
	ClearDNSCache();
	DeleteCriticalSection(&g_csDNS_Cache);
}

/*
 *	LbVɃf[^ǉ֐B
 *	
 *		szHost,addr ǉzXgIPAhX
 */
void AddDNSCache(char *szHost,unsigned long addr)
{
	struct DNS_CACHE *dns_new_cache;
	int n;

	InitializeDNSCache();
	EnterCriticalSection(&g_csDNS_Cache);
	if(g_dns_cache){
		if(g_nCacheCount + 1 >= g_nCacheSize){
			// LbVi[obt@ȂƂ
			// i[p̃obt@ĊmۂB
			n = g_nCacheSize + DNS_CACHE_SIZE;

			dns_new_cache = Malloc(n * sizeof(struct DNS_CACHE));
			CopyMemory(dns_new_cache,g_dns_cache,g_nCacheCount * sizeof(struct DNS_CACHE));
			GlobalFree(g_dns_cache);
			g_dns_cache = dns_new_cache;
			g_nCacheSize= n;
		}
		// LbVɃf[^ǉB
		g_dns_cache[g_nCacheCount].szHost = DupStr(szHost);
		g_dns_cache[g_nCacheCount].addr   = addr;
		g_nCacheCount++;
	}
	LeaveCriticalSection(&g_csDNS_Cache);
}

/*
 *	LbVIPAhXo֐B
 *	
 *		szHost	ׂzXg
 *	߂l
 *		LbVɃf[^΂IPAhXԂB
 *		f[^݂Ȃ-1ԂB
 */
unsigned long GetDNSCache(char *szHost)
{
	int i;
	unsigned long addr = (unsigned long)-1;

	InitializeDNSCache();
	EnterCriticalSection(&g_csDNS_Cache);
	if(g_dns_cache){
		for(i=0;i<g_nCacheCount;i++){
			if(lstrcmpi(g_dns_cache[i].szHost,szHost)==0){
				addr = g_dns_cache[i].addr;
				break;
			}
		}
	}
	LeaveCriticalSection(&g_csDNS_Cache);
	return addr;
}

int DebugPrint(char *format,...)
{
	va_list va_arg;
	char buf[8*1024];
	int n = 0;

	va_start(va_arg,format);
	vsprintf(buf,format,va_arg);
	va_end(va_arg);
#ifdef _CONSOLE
	WriteConsole(GetStdHandle(STD_OUTPUT_HANDLE),buf,lstrlen(buf),&n,NULL);
#endif
#ifdef _DEBUG
	OutputDebugString(buf);
#endif
	return n;
}

/*
 *	LbṼf[^\֐(fobOp)
 */
void DumpDNSCache(void)
{
	unsigned long addr;
	int i;

	InitializeDNSCache();
	EnterCriticalSection(&g_csDNS_Cache);
	if(g_dns_cache){
		for(i=0;i<g_nCacheCount;i++){
			addr = ntohl(g_dns_cache[i].addr);
			DebugPrint("%s %d.%d.%d.%d\n",g_dns_cache[i].szHost,
				   (addr >> 24) & 0xff,(addr >> 16) & 0xff,
				   (addr >> 8 ) & 0xff, addr & 0xff);
		}
	}
	LeaveCriticalSection(&g_csDNS_Cache);
}


/*
 *	zXgIPAhXo֐B
 *	
 *		szHost	ׂzXg
 *	߂l
 *		IPAhXԂBs-1ԂB
 */
unsigned long GetHostAddr(char *szHost)
{
	struct hostent *pHostEnt;
	unsigned long addr;

	// IPAhX\񂩂ǂׂ(: "192.168.0.1")
	addr = inet_addr(szHost);
	if(addr != -1){
		// LIPAhX̂Ƃ
		return addr;
	}

	// DNS̃LbVɃf[^邩ׂ
	addr = GetDNSCache(szHost);
	if(addr != -1){
		// LbVɃf[^݂
		return addr;
	}

	// LbVɑ݂Ȃ̂Ŗ₢킹
	if(addr == -1){
		// zXgnꂽƂ(szHostIPAhXłȂƂ)
		pHostEnt = gethostbyname(szHost);
		if(pHostEnt){
			addr = *((unsigned long *)((pHostEnt->h_addr_list)[0]));
		}
		AddDNSCache(szHost,addr);
	}
	return addr;
}
